/* Copyright (c)1994-2000 Begemot Computer Associates. All rights reserved.
 * See the file COPYRIGHT for details of redistribution and use. */
/* $Id: proc.h,v 1.11 2000/06/15 11:18:50 hbb Exp $ */

# include <sys/types.h>
# include <sys/param.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/mman.h>
# include <sys/stat.h>
# include <sys/wait.h>
# include <sys/uio.h>
# include <stdio.h>
# include <unistd.h>
# include <fcntl.h>
# include <signal.h>
# include <stdarg.h>
# ifdef HAVE_STREAMS
#  include <stropts.h>
# endif
# include <stdlib.h>
# include <string.h>
# include <setjmp.h>
# include <termios.h>
# include <errno.h>
# include <assert.h>
# include <limits.h>

# include <begemot.h>
# include "cdefs.h"
# include "pathnames.h"
# include "libutil/util.h"

/*
 * tunable parameters
 */
# define PHYSMEM 	(2*1024*1024)
# define FD_DEVICES	64
# define P11CONF	"p11conf"
# define MAXTIMEOUT	13

# ifndef HAVE_UCHAR
typedef unsigned char		uchar;
# endif
typedef signed char		schar;
typedef signed short		sshort;
typedef long long 		Quad;
typedef unsigned long long	UQuad;
typedef signed long long	SQuad;
# ifndef HAVE_UINT
typedef unsigned int		uint;
# endif
# ifndef HAVE_ULONG
typedef unsigned long	ulong;
# endif
typedef void		(*const VPF)();

typedef struct Proc 	Proc;
typedef struct MMG 	MMG;
typedef struct IOConf 	IOConf;
typedef struct IOOps  	IOOps;
typedef struct IODev  	IODev;
typedef struct IODevCmd	IODevCmd;
typedef struct Memops	Memops;
typedef struct FPRegs 	FPRegs;
typedef struct PDP_Float PDP_Float;
typedef enum DmaDir_t 	DmaDir_t;

typedef void	(*PFstore)(unsigned, int, int, int, ushort);
typedef unsigned(*PFfetch)(unsigned, int, int);

/*
 * entry points to non-generic memory part
 */
struct Memops {
	void	(*reset)(MMG *);		/* bus reset		*/
	void	(*mmr0)(MMG *);			/* mmr0 written		*/
	void	(*mmr3)(MMG *);			/* mmr3 written		*/
	void	(*par)(MMG *, int, int, int);	/* par written		*/
	void	(*pdr)(MMG *, int, int, int);	/* pdr written		*/
	void	(*ccr)(MMG *);			/* ccr written		*/
	void	(*info)(MMG *);			/* odt info		*/
	PFstore	store16, store18, store22;
	PFfetch	fetch16, fetch18, fetch22;
	ushort  *(*addrphys)(MMG *, unsigned);	/* for dma		*/
};

/*
 * floating status bit definitions
 */
enum {
	FER	= 0100000,	/* floating error */
	FID	=  040000,	/* interrupt disable */
	FIUV	=   04000,	/* interrupt on undefined variable */
	FIU	=   02000,	/* interrupt on underflow */
	FIV	=   01000,	/* interrupt on overflow */
	FIC	=    0400,	/* interrupt on integer conversion error */
	FD	=    0200,	/* floating double precision mode */
	FL	=    0100,	/* floating long-integer mode */
	FT	=     040,	/* floating chop mode */
	FN	=     010,	/* neg */
	FZ	=      04,	/* zero */
	FV	=      02,	/* overflow */
	FC	=      01,	/* carry */
};

/*
 * floating point errors
 */
enum {
	FE_ILL	= 2,		/* illegal opcode */
	FE_DIV0	= 4,		/* divide by zero */
	FE_INT	= 6,		/* integer conversion error */
	FE_OFL	= 8,		/* overflow */
	FE_UFL	= 10,		/* underflow */
	FE_UND	= 12,		/* undefined variable */
};

/*
 * pdp float/double format
 * this structure MUST be 4 ushorts (64 bits) long or some code will break
 */
# ifdef WORDS_BIGENDIAN
struct PDP_Float {
	unsigned	s : 1;		/* sign */
	unsigned	e : 8;		/* expo offset 0200 */
	unsigned	m0 : 7;		/* 7 bits in 1st word */
	unsigned	m1 : 16;	/* 1nd word */
	unsigned	m2 : 16;	/* 2nd word */
	unsigned	m3 : 16;	/* 3nd word */
};
# else
struct PDP_Float {
	unsigned	m0 : 7;		/* 7 bits in 1st word */
	unsigned	e : 8;		/* expo offset 0200 */
	unsigned	s : 1;		/* sign */
	unsigned	m1 : 16;	/* 1nd word */
	unsigned	m2 : 16;	/* 2nd word */
	unsigned	m3 : 16;	/* 3nd word */
};
# endif
# define PDP_OFF	0200		/* exponent excess */

/*
 * floating point processor
 */
struct FPRegs {
	PDP_Float	ac[6];		/* AC0-6 */
	ushort		st;		/* state */
	ushort		fec;		/* exception code */
	ushort		fea;		/* exception address */
};


/*
 * processor
 */
struct Proc {
	ushort		reg[8];		/* registers for cur. mode	*/
	ushort		ureg[6];	/* R0-R5			*/
	ushort		kreg[6];	/* R0'-R5'			*/
	ushort		sp[4];		/* SP				*/

	ushort		cpuerr;		/* CPU error register		*/
	ushort		pirq;		/* programmed interrupt request	*/
	ushort		ltc;		/* line time clock		*/
	ushort		maint;		/* maintenance			*/
	ushort		ccr;		/* cache control :-)		*/
	int		bevent;		/* if true - timer disabled 	*/

	int		fpd;		/* disable FPP			*/
	FPRegs		fpp;		/* floating point registers	*/

	/* PSW */
	uchar		c, n, v, z;
	int		t;
	int		pri;
	int		regs;
	int		curmode;	/* mode from psw 		*/
	int		prevmode;	/* previous mode  		*/

	MMG		*mmg;		/* memory manager		*/
	Memops		*memops;	/* operations			*/
	ushort		mmr0;
	unsigned	mmr2;		/* trapping pc			*/
	unsigned	mmr3;		/* 				*/
	int		mmfrozen;
	ushort		fmmr0, fmmr1, fmmr2;
	int		mmr1reg[2];	/* register part		*/
	int		mmr1chg[2];	/* change part			*/
	PFfetch 	fetch;
	PFstore		store;
	ushort		par[4][2][8];
	ushort		pdr[4][2][8];

	unsigned 	physmem;	/* amount of phys. mem (bytes)	*/
	IODev		*iopage[4096];

	int		reqpri;		/* highest pending bus request	*/
	IODev		*devices;
	IODev		*proc;		/* processor registers */

	unsigned	cmd;		/* last ifetch			*/
	int		push;		/* if true last op did stack push */
	ushort		bp;		/* breakpoint 			*/
	int		tinhibit;	/* tbit trap pending		*/
	int		halt;		/* stop processor		*/
};

/*
 * "Bootstrap" structure.
 */
struct IOConf {
	char	*name;			/* name ala "kl", "rl", ...	*/
	IOOps	*ioops;			/* class specific functions	*/
	int	internal;		/* not really a device 		*/
};

/*
 * Per-Controller-Class structure
 */
struct IOOps {
	void		(*ctrl_create)(IODev *,int, char **);
	void		(*dev_create)(IODev *,int, char **);
	void		(*ctrl_complete)(IODev *);
	void		(*reset)(IODev *);
	ushort		(*fetch)(IODev *, unsigned);
	void		(*store)(IODev *, unsigned, int, ushort);
	ushort		(*vector)(IODev *);
	void		(*dma)(IODev *);
	void		(*async)(IODev *, void *);
	void		(*info)(IODev *);
	void		(*flush)(IODev *);

	IODevCmd	*cmds;
};

/*
 * Per-Controller-structure
 */
struct IODev {
	int		reqpri;
	IODev		*next;
	void		*data;
	char		*name;
	IOOps		*ioops;
	int		unit;
};

/*
 * device command table entry
 */
struct IODevCmd {
	char		*name;
	char		*usage;
	int		(*func)(IODev *, int, char **);
};


void	dev_command(IODev *, int, char **);
int	dev_help(IODev *, int, char **);

enum DmaDir_t {
	WriteMem,	/* from buffer into memory */
	ReadMem,	/* from memory to buffer */
	CmpMem,		/* compare memory and buffer */
};
uint	dma(void *buf, uint bytes, uint mem, DmaDir_t direction, int *);

extern Proc 	proc;
extern ushort	startloc;
extern int	ill_trace;
extern int	nomon;
extern char	*libdir;
extern char	*iodir;
extern uint	clock_rate;
extern uint	csw_enable;
extern uint	csw;

/*
 * CPU options/maintenance register bits
 */
extern uint	cpu_options;
enum {
	CPU_PWOK	= 0000001,	/* BPOK (AC power ok) */
	CPU_PWUP_MASK	= 0000006,	/* power up mask */
	 CPU_PWUP_T24	= 0000000,	/* trap to 24 */
	 CPU_PWUP_MON	= 0000002,	/* enter monitor */
	 CPU_PWUP_AUTO	= 0000004,	/* 173000g */
	 CPU_PWUP_USER	= 0000006,	/* user boot address (from high bits) */
	CPU_HALT	= 0000010,	/* trap to 4 on kernel HALT */
	CPU_MOD_MASK	= 0000360,	/* module type */
	 CPU_MOD_KDJ11A	= 0000020,	/* KDJ11A dual board 11/73 */
	 CPU_MOD_KDJ11B	= 0000040,	/* KDJ11B quad board 11/73,83,84 */
	 CPU_MOD_KXJ11C	= 0000060,	/* KDJ11X co-processor */
	 CPU_MOD_KDJ11D	= 0000100,	/* KDJ11D 11/53 */
	 CPU_MOD_KDJ11E	= 0000120,	/* KDJ11E 11/93,94 */
	CPU_FPA		= 0000400,	/* FPA available */
	CPU_UNIBUS	= 0001000,	/* UNIBUS adaptor available */
	CPU_USER	= 0176000,	/* user start address */
};

/*
 * Memory management bits
 */
enum {
	MMR0_ANR	= 0100000,	/* abort non-resident */
	MMR0_APL	= 0040000,	/* abort page length */
	MMR0_ARO	= 0020000,	/* abort read-only */
	MMR0_ABORT_MASK	= 0160000,	/* all abort bits */
	MMR0_MAI	= 0000400,	/* maintenance (not implemented) */
	MMR0_MODE	= 0000140,	/* mode from PSW at during abort */
	MMR0_MODE_SHIFT	= 5,
	MMR0_SPACE	= 0000020,	/* address space during abort */
	MMR0_SPACE_SHIFT= 4,
	MMR0_PAGE	= 0000016,	/* aborted page number */
	MMR0_PAGE_SHIFT	= 1,
	MMR0_ENABLE	= 0000001,	/* memory management enable */

	MMR3_IOM	= 0000040,	/* not implemented */
	MMR3_22B	= 0000020,	/* enable 22 bit mapping */
	MMR3_KSEP	= 0000004,	/* kernel separate I/D enable */
	MMR3_SSEP	= 0000002,	/* supervisor separate I/D enable */
	MMR3_USEP	= 0000001,	/* user separate I/D enable */
	MMR3_MASK	= 0000077,	/* writeable bits */
};
char *	print_mmr0(ushort);
char *	print_mmr1(ushort);
char *	print_mmr3(ushort);

extern char	verbose[128];

extern IOOps	rl_ops;
extern IOOps	kl_ops;
extern IOOps	mr_ops;
extern IOOps	qna_ops;
extern IOOps	memsys_ops;
extern IOOps	proc_ops;

extern jmp_buf	trap;

void	push(ushort);
ushort	pop();
void	switchmode(int, int);
void	interp();
void	Reset();
void	Flush();
void	catch_io_sigs();

/*
 * generic memory interface
 */
ushort 	iofetch(unsigned);
void 	iostore(unsigned, int, ushort);
# define mmfetch(A,D,M)		proc.fetch((A),(D),(M))
# define mmstore(A,T,D,M,V)	proc.store((A),(T),(D),(M),(V))
# define mfetch(A,D) 		mmfetch((A),(D),proc.curmode)
# define mstore(A,T,D,V)	mmstore((A),(T),(D),proc.curmode,(V))
void	mem_init();		/* 1st initialisation	*/


volatile void	Trap(ushort) DEAD_FUNC;
volatile void	Trap10() DEAD_FUNC;
volatile void	Trap4(ushort) DEAD_FUNC;
volatile void	YellowTrap() DEAD_FUNC;
volatile void 	MemTrap(int, int, int, ushort) DEAD_FUNC;
void		SetupTrap(ushort);
void		proc_pwrfail(int real);
void		pwrfail_exit();

void		mon_fpp();
void		store_long(int);
void		store_float(ushort *, int, int);
void 		load_float(ushort *, int, int);
void		load_long(int);

unsigned	parse_csr(char *, char *);
ushort		parse_vec(char *, char *);
ushort		parse_irq(char *, char *);
int		getmfields(char *, char **, int);
char		*setfields(char *);

int		register_handler(char *, char *[], IODev *, void *, int, pid_t *);

int		register_timer(unsigned, void (*)(void *), void *);
void		reset_timer(unsigned, int);

void		monitor();
void		type_help_file(char *help);
void		request();
void		conf_warning(char *, ...);
volatile void	conf_panic(char *, ...) DEAD_FUNC;

void		init_timer();
void		start_timer();
int		stop_timer();

enum {
	M_Mode		= 03,	/* mode mask			*/
	M_Low		= 010,	/* low byte			*/
	M_High		= 020,	/* high byte			*/
	M_Word		= 040,	/* word				*/
};

# define Max(A,B)	(((A) > (B)) ? (A) : (B))
# define Off(A) ((A) >> 1)
# define IOP(A) (((A) & 017777) >> 1)

/*
 * host byte ordering dependend stuff
 */
# ifdef WORDS_BIGENDIAN
#  define SETLB(W, V)    ((W) = ((W) & 0377) | ((V) << 8 ))
#  define SETHB(W, V)    ((W) = ((W) & 0177400) | ((V) >> 8))
#  define SETW(W,V)      ((W) = ((V)<<8) | (((V)>>8)&0xff))
#  define GETW(V)        ((((V)&0xff)<<8)|(((V)>>8)&0xff))
# else
#  define SETLB(W, V)    ((W) = ((W) & 0177400) | ((V) & 0377))
#  define SETHB(W, V)    ((W) = ((W) & 0377) | ((V) & 0177400))
#  define SETW(W,V)      ((W) = (V))
#  define GETW(V)        (V)
# endif
# define SLB(W, V)    ((W) = ((W) & 0177400) | ((V) & 0377))
# define SHB(W, V)    ((W) = ((W) & 0377) | ((V) & 0177400))

/*
 * Must be higher, than anything else ...
 */
# define NMIPRI 	10
# define DMAPRI		11
# define IRQ(D, L)	((D)->reqpri = (L), proc.reqpri = Max(proc.reqpri,(L)))

# define SUBSTR(SUB,S) (strlen((SUB)) <= strlen((S)) && strncmp((S), (SUB), strlen((SUB))) == 0)
# define xstrdup(S,F)	strcpy(xalloc(strlen((S))+1, (F)), (S))

# ifndef MAP_FILE
#  define MAP_FILE 0
# endif

# ifndef HAVE_BZERO
#  define bzero(P,L) (void)memset((P),0,(L))
# endif

# ifndef HAVE_BCOPY
#  define bcopy(F,T,L) (void)memcpy((T),(F),(L))
# endif
